<?php

namespace UnicaenDbImport\CodeGenerator;

use Doctrine\DBAL\DBALException;
use Doctrine\DBAL\Platforms\AbstractPlatform;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\CodeGenerator\Helper\DiffViewHelper;
use UnicaenDbImport\CodeGenerator\Helper\LogTableHelper;
use UnicaenDbImport\CodeGenerator\Helper\TableHelper;
use UnicaenDbImport\CodeGenerator\Helper\TableValidationHelper;
use UnicaenDbImport\Domain\DestinationInterface;
use UnicaenDbImport\Domain\ResultInterface;
use UnicaenDbImport\Domain\SourceInterface;

/**
 * Classe mère des générateurs de code SQL, quelque soit la plateforme de base de données.
 *
 * @author Unicaen
 */
abstract class CodeGenerator implements CodeGeneratorInterface
{
    /**
     * @var AbstractPlatform
     */
    protected $platform;

    /**
     * @var TableHelper
     */
    protected $tableHelper;

    /**
     * @var TableValidationHelper
     */
    protected $tableValidationHelper;

    /**
     * @var DiffViewHelper
     */
    protected $diffViewHelper;

    /**
     * @var LogTableHelper
     */
    protected $logTableHelper;

    /**
     * AbstractCodeGenerator constructor.
     *
     * @param AbstractPlatform $platform
     */
    public function __construct(AbstractPlatform $platform)
    {
        $this->platform = $platform;

        //
        // NB: Les classes filles ont le responsabilité d'instancier les helpers concrets.
        //
    }

    /**
     * @param string $tableName
     * @return string
     * @codeCoverageIgnore Car simple délégation
     */
    public function generateSQLForTableExistenceCheck($tableName)
    {
        return $this->tableValidationHelper->generateSQLForTableExistenceCheck($tableName);
    }

    /**
     * @param array $result
     * @return bool
     * @codeCoverageIgnore Car simple délégation
     */
    public function convertTableExistenceCheckResultToBoolean(array $result)
    {
        return $this->tableValidationHelper->convertTableExistenceCheckResultToBoolean($result);
    }

    /**
     * @param string $tableName
     * @param array  $columns
     * @return string
     * @codeCoverageIgnore Car simple délégation
     */
    public function generateSQLForDataColumnsValidation($tableName, array $columns)
    {
        return $this->tableValidationHelper->generateSQLForColumnsValidation($tableName, $columns);
    }

    /**
     * @param string $tableName
     * @param array  $result
     * @param array  $expectedColumns
     * @return null|RuntimeException
     * @codeCoverageIgnore Car simple délégation
     */
    public function convertDataColumnsValidationBadResultToException($tableName, array $result, array $expectedColumns)
    {
        return $this->tableValidationHelper->convertColumnsValidationResultToException($tableName, $result, $expectedColumns);
    }

    /**
     * @param string $tableName
     * @return string
     * @codeCoverageIgnore Car simple délégation
     */
    public function generateSQLForHistoColumnsValidation($tableName)
    {
        return $this->tableValidationHelper->generateSQLForHistoColumnsValidation($tableName);
    }

    /**
     * @param string $tableName
     * @param array  $result
     * @return null|RuntimeException
     * @codeCoverageIgnore Car simple délégation
     */
    public function convertHistoColumnsValidationBadResultToException($tableName, array $result)
    {
        return $this->tableValidationHelper->convertHistoColumnsValidationBadResultToException($tableName, $result);
    }

    /**
     * @param string $tableName
     * @return string
     * @codeCoverageIgnore Car simple délégation
     */
    public function generateSQLForSourceColumnsValidation($tableName)
    {
        return $this->tableValidationHelper->generateSQLForSourceColumnsValidation($tableName);
    }

    /**
     * @param string $tableName
     * @param array  $result
     * @return null|RuntimeException
     * @codeCoverageIgnore Car simple délégation
     */
    public function convertSourceColumnsValidationBadResultToException($tableName, array $result)
    {
        return $this->tableValidationHelper->convertSourceColumnsValidationBadResultToException($tableName, $result);
    }

    /**
     * @param SourceInterface $source
     * @return string
     */
    public function generateSQLForSelectFromSource(SourceInterface $source)
    {
        if ($source->getSelect()) {
            return $source->getSelect();
        }

        $sourceTable = $source->getTable();

        return $this->tableHelper->generateSQLForSelectFromTable($sourceTable);
    }

    /**
     * @param SourceInterface      $source
     * @param DestinationInterface $destination
     * @param string               $intermediateTable
     * @return string
     */
    public function generateSQLForIntermediateTableCreation(SourceInterface $source, DestinationInterface $destination, $intermediateTable)
    {
        $destinationTable = $destination->getTable();
        $sourceCodeColumn = $source->getSourceCodeColumn();
        $columns = $source->getColumns();

        return $this->tableHelper->generateSQLForIntermediateTableCreation($destinationTable, $sourceCodeColumn, $columns, $intermediateTable);
    }

    /**
     * @param string $tableName
     * @param array  $columnsValues
     * @return string
     */
    public function generateSQLForInsertIntoTable($tableName, array $columnsValues)
    {
        return $this->tableHelper->generateSQLForInsertIntoTable($tableName, $columnsValues);
    }

    /**
     * @param string $tableName
     * @return string
     */
    public function generateSQLForTruncateTable($tableName)
    {
        $platform = $this->platform;

        return $platform->getTruncateTableSQL($tableName);
    }

    /**
     * @param string $logTableName
     * @return string
     * @codeCoverageIgnore
     */
    public function generateSQLForImportLogTableCreation(string $logTableName)
    {
        return $this->logTableHelper->generateSQLForLogTableCreation($logTableName);
    }

    /**
     * @param string $tableName
     * @return string
     */
    public function generateSQLForIntermmediateTableDrop($tableName)
    {
        throw CodeGeneratorException::methodToImplement(__METHOD__, get_called_class(), $this->platform);
    }

    /**
     * @param SourceInterface      $source
     * @param DestinationInterface $destination
     * @param string               $intermediateTable
     * @return string
     * @throws DBALException
     */
    public function generateSQLForDiffViewCreation(SourceInterface $source, DestinationInterface $destination, $intermediateTable)
    {
        $destinationTable = $destination->getTable();
        $sourceCodeColumn = $source->getSourceCodeColumn();
        $columns = $source->getColumns();

        $diffSourceTableName = $intermediateTable ?: $source->getTable();

        return $this->diffViewHelper->generateSQLForCreation($diffSourceTableName, $destinationTable, $sourceCodeColumn, $columns);
    }

    /**
     * @param SourceInterface      $source
     * @param DestinationInterface $destination
     * @return string
     */
    public function generateSQLForDiffViewSelect(SourceInterface $source, DestinationInterface $destination)
    {
        $destinationTable = $destination->getTable();
        $sourceCodeColumn = $source->getSourceCodeColumn();
        $columns = $source->getColumns();

        return $this->diffViewHelper->generateSQLForSelect($destinationTable, $sourceCodeColumn, $columns);
    }

    /**
     * @param array                $rows
     * @param SourceInterface      $source
     * @param DestinationInterface $destination
     * @return string
     */
    public function generateSQLForDestinationUpdateFromDiffViewRequestResult(
        array $rows,
        SourceInterface $source,
        DestinationInterface $destination)
    {
        $destinationTable = $destination->getTable();
        $sourceCodeColumn = $source->getSourceCodeColumn();
        $columns = $source->getColumns();

        return $this->diffViewHelper->generateSQLForOperationInDestinationTableFromDiffViewRequestResult($rows, $destinationTable, $sourceCodeColumn, $columns);
    }

    /**
     * @param ResultInterface $result
     * @param string          $importLogTable
     * @return string
     */
    public function generateSQLForInsertResultIntoLogTable(ResultInterface $result, string $importLogTable)
    {
        return $this->logTableHelper->generateSQLForInsertResultIntoLogTable($result, $importLogTable);

    }
}