<?php

namespace UnicaenDbImport\CodeGenerator\Helper;

use Doctrine\DBAL\DBALException;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\CodeGenerator\Helper;
use UnicaenDbImport\Domain\Operation;

/**
 * Génération de code SQL de création de la vue "v_diff_{table}".
 *
 * Version commune à toutes les plateformes de bases de données.
 *
 * @author Unicaen
 */
abstract class DiffViewHelper extends Helper
{
    /**
     * @param string $diffSourceTableName
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     * @throws DBALException
     */
    public function generateSQLForCreation($diffSourceTableName, $destinationTable, $sourceCodeColumn, array $columns)
    {
        $res = '';
        $res .= $this->generateViewDeletionSQLSnippet($destinationTable) . ';' . PHP_EOL;
        $res .= $this->generateViewCreationSQLSnippet($diffSourceTableName, $destinationTable, $sourceCodeColumn, $columns) . PHP_EOL;

        return $res;
    }

    /**
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     */
    public function generateSQLForSelect($destinationTable, $sourceCodeColumn, array $columns)
    {
        $name = $this->generateViewName($destinationTable);

        $updatedColumsSql = implode(
            array_map(function($col) {
                return sprintf('U_%s', $col);
            }, $columns),
            ','
        );

        $srcColumsSql = implode(
            array_map(function($col) {
                return sprintf('S_%s', $col);
            }, $columns),
            ',' . PHP_EOL
        );

        $destColumsSql = implode(
            array_map(function($col) {
                return sprintf('D_%s', $col);
            }, $columns),
            ',' . PHP_EOL
        );

        $sql = <<<EOT
SELECT 
    $sourceCodeColumn, 
    operation, 
    $updatedColumsSql, 
    $srcColumsSql, 
    $destColumsSql 
FROM $name
EOT;

        return $sql;
    }

    /**
     * @param string $destinationTable
     * @return string
     */
    protected function generateViewName($destinationTable)
    {
        return 'V_DIFF_' . $destinationTable;
    }

    /**
     * @param string $destinationTable
     * @return string
     * @throws DBALException
     */
    protected function generateViewDeletionSQLSnippet($destinationTable)
    {
        $name = $this->generateViewName($destinationTable);

        return $this->platform->getDropViewSQL($name);
    }

    /**
     * @param string $diffSourceTableName
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     */
    protected function generateViewCreationSQLSnippet($diffSourceTableName, $destinationTable, $sourceCodeColumn, array $columns)
    {
        $name = $this->generateViewName($destinationTable);

        $res = '';
        $res .= "CREATE VIEW $name AS " . PHP_EOL;
        $res .= $this->indent(4, $this->generateViewBodySelectSQLSnippet($diffSourceTableName, $destinationTable, $sourceCodeColumn, $columns)) . PHP_EOL;

        return $res;
    }

    /**
     * @param string $diffSourceTableName
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     */
    protected function generateViewBodySelectSQLSnippet($diffSourceTableName, $destinationTable, $sourceCodeColumn, array $columns)
    {
        $insert = Operation::OPERATION_INSERT;
        $update = Operation::OPERATION_UPDATE;
        $undelete = Operation::OPERATION_UNDELETE;
        $delete = Operation::OPERATION_DELETE;

        $updatedColumsSql = implode(
            array_map(function($col) {
                $tpl = 'CASE WHEN d.%s <> s.%1$s OR (d.%1$s IS NULL AND s.%1$s IS NOT NULL) OR (d.%1$s IS NOT NULL AND s.%1$s IS NULL) THEN 1 ELSE 0 END as U_%1$s';
                return sprintf($tpl, $col);
            }, $columns),
            ',' . PHP_EOL
        );

        $srcColumsSql = implode(
            array_map(function($col) {
                return sprintf('s.%1$s AS S_%1$s', $col);
            }, $columns),
            ',' . PHP_EOL
        );

        $destColumsSql = implode(
            array_map(function($col) {
                return sprintf('d.%1$s AS D_%1$s', $col);
            }, $columns),
            ',' . PHP_EOL
        );

        $udpatedColumnsAddition = implode(
            array_map(function($col) {
                return sprintf('U_%s', $col);
            }, $columns),
            '+'
        );

        $sql = <<<EOT
with diff as (
    SELECT
        coalesce(s.$sourceCodeColumn, d.$sourceCodeColumn) $sourceCodeColumn,

        case 
        when (s.$sourceCodeColumn IS NOT NULL AND d.$sourceCodeColumn IS NULL) then '$insert'
        when (s.$sourceCodeColumn IS NOT NULL AND d.$sourceCodeColumn IS NOT NULL and (d.deleted_on IS NULL or d.deleted_on > now())) then '$update'
        when (s.$sourceCodeColumn IS NOT NULL AND d.$sourceCodeColumn IS NOT NULL and (d.deleted_on IS NOT NULL and d.deleted_on <= now())) then '$undelete'
        when (s.$sourceCodeColumn IS NULL AND d.$sourceCodeColumn IS NOT NULL and (d.deleted_on IS NULL or d.deleted_on > now())) then '$delete' end as operation,

        $updatedColumsSql,

        $srcColumsSql,

        $destColumsSql,
        
        s.source_id

    FROM $destinationTable d
    JOIN source src ON src.id = d.source_id AND src.synchronisable = 1
    FULL OUTER JOIN $diffSourceTableName s ON s.$sourceCodeColumn = d.$sourceCodeColumn
)
select * from diff
where operation is not null
and (operation = '$undelete' or 0 < $udpatedColumnsAddition)
EOT;

        return $sql;
    }

    /**
     * @param array    $rows
     * @param string   $destinationTable
     * @param string   $sourceCodeColumn
     * @param string[] $columns
     * @return string
     */
    public function generateSQLForOperationInDestinationTableFromDiffViewRequestResult(array $rows, $destinationTable, $sourceCodeColumn, $columns)
    {
        // réorganisation par type d'opération
        $data = [];
        foreach ($rows as $row) {
            $operation = $row['operation'];
            if (!isset($data[$operation])) {
                $data[$operation] = [];
            }
            $data[$operation][] = $row;
        };

        $sqls = [];

        // une instruction SQL (update ou insert) par type d'opération suffira à mettre à jour la table destination
        $operations = array_keys($data);
        foreach ($operations as $operation) {
            switch ($operation) {
                case Operation::OPERATION_INSERT:
                    $sqls[] = $this->generateSQLForInsertOperationInDestinationTable($destinationTable, $sourceCodeColumn, $columns) . ';' . PHP_EOL;
                    break;
                case Operation::OPERATION_UPDATE:
                    $sqls[] = $this->generateSQLForUpdateOperationInDestinationTable($destinationTable, $sourceCodeColumn, $columns) . ';' . PHP_EOL;
                    break;
                case Operation::OPERATION_UNDELETE:
                    $sqls[] = $this->generateSQLForUndeleteOperationInDestinationTable($destinationTable, $sourceCodeColumn, $columns) . ';' . PHP_EOL;
                    break;
                case Operation::OPERATION_DELETE:
                    $sqls[] = $this->generateSQLForDeleteOperationInDestinationTable($destinationTable, $sourceCodeColumn, $columns) . ';' . PHP_EOL;
                    break;
                default:
                    throw new RuntimeException("Opération inconnue rencontrée : " . $operation);
                    break;
            }
        }

        return empty($sqls) ? '' : implode(PHP_EOL, $sqls);
    }

    /**
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array  $columns
     * @return string
     */
    protected function generateSQLForInsertOperationInDestinationTable($destinationTable, $sourceCodeColumn, array $columns)
    {
        $diffViewName = $this->generateViewName($destinationTable);

        $now = $this->platform->getNowExpression();

        $cols = array_merge([$sourceCodeColumn], $columns);
        $sourceCols = array_map(function ($col) {
            return 'S_' . $col;
        }, $columns);

        $insert = Operation::OPERATION_INSERT;

        // instruction de mise à jour de la table destination
        $markup = '';
        $markup .= "INSERT INTO $destinationTable (created_on, " . implode(', ', $cols) . ')' . PHP_EOL;
        $markup .= "SELECT $now, $sourceCodeColumn, " . implode(', ', $sourceCols) . PHP_EOL;
        $markup .= "FROM $diffViewName WHERE operation = '$insert' ";
        
        return $markup;
    }

    /**
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array $columns
     * @return string
     */
    abstract protected function generateSQLForUpdateOperationInDestinationTable($destinationTable, $sourceCodeColumn, array $columns);

    /**
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array $columns
     * @return string
     */
    abstract protected function generateSQLForUndeleteOperationInDestinationTable($destinationTable, $sourceCodeColumn, array $columns);

    /**
     * @param string $destinationTable
     * @param string $sourceCodeColumn
     * @param array $columns
     * @return string
     */
    abstract protected function generateSQLForDeleteOperationInDestinationTable($destinationTable, $sourceCodeColumn, array $columns);
}