<?php

namespace UnicaenDbImport\Config;

use Assert\Assertion;
use Assert\AssertionFailedException;
use Doctrine\DBAL\Connection as DbConnection;
use Interop\Container\ContainerInterface;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\Connection\ApiConnection;
use UnicaenDbImport\Domain\ColumnNameFilter;
use UnicaenDbImport\Domain\Destination;
use UnicaenDbImport\Domain\Import;
use UnicaenDbImport\Domain\Source;
use UnicaenDbImport\Domain\Synchro;
use Zend\Filter\FilterInterface;
use Zend\ServiceManager\ServiceLocatorInterface;

class ConfigFactory
{
    const CONFIG_KEY = 'import';

    /**
     * @var ServiceLocatorInterface
     */
    private $container;

    /**
     * @var array
     */
    private $moduleConfig;

    /**
     * @var array
     */
    private $connections;

    /**
     * @param ContainerInterface $container
     * @return Config
     */
    public function __invoke(ContainerInterface $container)
    {
        $this->container = $container;
        $this->moduleConfig = $this->getModuleConfig();

        $codeGeneratorsConfig = $this->getCodeGeneratorsConfig();
        $imports = $this->createImports();
        $synchros = $this->createSynchros();

        $config = new Config();
        $config->setCodeGeneratorsConfig($codeGeneratorsConfig);
        $config->setImports($imports);
        $config->setSynchros($synchros);

        return $config;
    }

    /**
     * @return mixed
     */
    private function getModuleConfig()
    {
        $appConfig = $this->container->get('Config');

        if (!array_key_exists($key = self::CONFIG_KEY, $appConfig)) {
            throw ConfigException::missingKey($key);
        }

        return $appConfig[$key];
    }

    /**
     * @param
     */
    private function createConnections()
    {
        try {
            Assertion::keyExists($this->moduleConfig, $key = 'connections');
        } catch (AssertionFailedException $e) {
            throw ConfigException::missingKey($key);
        }
        try {
            Assertion::isArray($this->moduleConfig[$key]);
        } catch (AssertionFailedException $e) {
            throw ConfigException::invalidValueForKey($key, 'array');
        }

        $connections = [];
        foreach ($this->moduleConfig['connections'] as $key => $value) {
            $connections[$key] = $this->createConnection($value);
        }

        $this->connections = $connections;
    }

    /**
     * @param string|array $value
     * @return DbConnection|ApiConnection
     */
    private function createConnection($value)
    {
        if (!is_string($value) && !is_array($value)) {
            throw new ConfigException("Une connection ne peut être créée qu'à partir d'une string ou d'un array");
        }

        if (is_string($value)) {
            $connectionName = $value;
            /** @var DbConnection $conn */
            $conn = $this->container->get($connectionName);
        } else {
            $conn = ApiConnection::fromArrayConfig($value);
        }

        return $conn;
    }

    /**
     * @return Import[]
     */
    private function createImports()
    {
        $imports = [];
        if (array_key_exists($key = 'imports', $this->moduleConfig)) {
            foreach ($this->moduleConfig[$key] as $importArr) {
                $imports[] = $this->createImport($importArr);
            }
        }

        return $imports;
    }

    /**
     * @param array $importConfig
     * @return Import
     */
    private function createImport(array $importConfig)
    {
        if (!array_key_exists($key = 'source', $importConfig)) {
            throw ConfigException::missingKey($key);
        }
        if (!array_key_exists($key = 'destination', $importConfig)) {
            throw ConfigException::missingKey($key);
        }

        $source = $this->createSource($importConfig['source']);
        $destination = $this->createDestination($importConfig['destination']);

        $importConfig['source'] = $source;
        $importConfig['destination'] = $destination;

        return Import::fromConfig($importConfig);
    }

    /**
     * @return Synchro[]
     */
    private function createSynchros()
    {
        $synchros = [];
        if (array_key_exists($key = 'synchros', $this->moduleConfig)) {
            foreach ($this->moduleConfig[$key] as $synchroArr) {
                $synchros[] = $this->createSynchro($synchroArr);
            }
        }

        return $synchros;
    }

    /**
     * @param array $synchroConfig
     * @return Synchro
     */
    private function createSynchro(array $synchroConfig)
    {
        if (!array_key_exists($key = 'source', $synchroConfig)) {
            throw ConfigException::missingKey($key);
        }
        if (!array_key_exists($key = 'destination', $synchroConfig)) {
            throw ConfigException::missingKey($key);
        }

        $source = $this->createSource($synchroConfig['source']);
        $destination = $this->createDestination($synchroConfig['destination']);

        $synchroConfig['source'] = $source;
        $synchroConfig['destination'] = $destination;

        return Synchro::fromConfig($synchroConfig);
    }

    /**
     * @param array $sourceConfig
     * @return Source
     */
    private function createSource(array $sourceConfig)
    {
        $connectionName = $sourceConfig['connection'];
        $sourceConfig['connection'] = $this->getConnectionByName($connectionName);

        $columnNameFilter = $this->getColumnNameFilter();
        $sourceConfig['column_name_filter'] = $columnNameFilter;

        $source = Source::fromConfig($sourceConfig);

        return $source;
    }

    /**
     * @param array $destinationConfig
     * @return Destination
     */
    private function createDestination(array $destinationConfig)
    {
        $connectionName = $destinationConfig['connection'];
        $destinationConfig['connection'] = $this->getConnectionByName($connectionName);

        return Destination::fromConfig($destinationConfig);
    }

    /**
     * @param string $connectionName
     * @return DbConnection|ApiConnection
     */
    private function getConnectionByName($connectionName)
    {
        if ($this->connections === null) {
            $this->createConnections();
        }

        try {
            Assertion::keyExists($this->connections, $connectionName);
        } catch (AssertionFailedException $e) {
            throw ConfigException::missingKey($connectionName);
        }
        /** @var DbConnection|ApiConnection $conn */
        $conn = $this->connections[$connectionName];

        return $conn;
    }

    /**
     * @return array
     */
    private function getCodeGeneratorsConfig()
    {
        try {
            Assertion::keyExists($this->moduleConfig, $key = 'code_generators');
        } catch (AssertionFailedException $e) {
            throw ConfigException::missingKey($key);
        }
        try {
            Assertion::isArray($this->moduleConfig[$key]);
        } catch (AssertionFailedException $e) {
            throw ConfigException::invalidValueForKey($key, 'array');
        }

        return $this->moduleConfig[$key];
    }

    /**
     * @return FilterInterface
     */
    private function getColumnNameFilter()
    {
        $cnf = $this->moduleConfig['column_name_filter'] ?? ColumnNameFilter::class;

        $filterInstance = NULL;

        if (is_string($cnf)) {
            if ($this->container->has($cnf)) {
                /** @var FilterInterface $filterInstance */
                $filterInstance = $this->container->get($cnf);
            } elseif (class_exists($cnf)) {
                $filterInstance = new $cnf;
            }
        } elseif ($cnf instanceof FilterInterface) {
            $filterInstance = $cnf;
        }

        if ($filterInstance === NULL) {
            throw ConfigException::invalidValueForKey('column_name_filter');
        }

        return $filterInstance;
    }
}