<?php

namespace UnicaenDbImport\Controller;

use DateTime;
use Exception;
use UnicaenDbImport\Domain\Exception\NotFoundException;
use UnicaenDbImport\Domain\ImportInterface;
use UnicaenDbImport\Domain\SynchroInterface;
use UnicaenDbImport\Service\ImportService;
use UnicaenDbImport\Service\SynchroService;
use Zend\Console\Charset\Utf8;
use Zend\Console\ColorInterface;
use Zend\Console\Request;
use Zend\Mvc\Console\Controller\AbstractConsoleController;

class ConsoleController extends AbstractConsoleController
{
    /**
     * @var ImportService
     */
    private $importService;

    /**
     * @param ImportService $importService
     */
    public function setImportService(ImportService $importService)
    {
        $this->importService = $importService;
    }

    /**
     * @var SynchroService
     */
    private $synchroService;

    /**
     * @param SynchroService $synchroService
     */
    public function setSynchroService(SynchroService $synchroService)
    {
        $this->synchroService = $synchroService;
    }

    /**
     * @param ImportInterface[] $imports
     */
    private function runImports($imports)
    {
        if (count($imports) === 0) {
            $this->getConsole()->writeLine("Aucun import réalisé.", ColorInterface::NORMAL);

            return;
        }

        foreach ($imports as $import) {

            $this->getConsole()->writeLine("### " . $import->getName() . " ###", ColorInterface::BLACK, ColorInterface::WHITE);
            $this->getConsole()->writeLine(date_format(date_create(), 'd/m/Y H:i:s'), ColorInterface::NORMAL);
            $this->getConsole()->writeLine();

            $result = $this->importService->runImport($import);

            if ($exception = $result->getException()) {
                $this->getConsole()->writeLine("Une erreur a été rencontrée!", ColorInterface::RED);
                $this->getConsole()->writeLine($exception->getMessage(), ColorInterface::RED);
                if ($previous = $exception->getPrevious()) {
                    $this->getConsole()->writeLine($previous->getMessage(), ColorInterface::RED);
                }
                $this->getConsole()->writeLine($exception->getTraceAsString(), ColorInterface::GRAY);
            }
            else {
                $this->getConsole()->writeLine((string)$result);
                $this->getConsole()->writeLine("Import réalisé avec succès.", ColorInterface::GREEN);
            }

            $startDate = $result->getStartDate();
            $endDate = $result->getEndDate();
            $diffDate = date_diff($startDate, $endDate);

            $this->getConsole()->writeLine($diffDate->format('%i min %s sec.'), ColorInterface::NORMAL);
            $this->getConsole()->writeLine();
        }
    }

    /**
     * @param SynchroInterface[] $synchros
     */
    private function runSynchros($synchros)
    {
        if (count($synchros) === 0) {
            $this->getConsole()->writeLine("Aucune synchro réalisée.", ColorInterface::NORMAL);

            return;
        }

        foreach ($synchros as $synchro) {

            $this->getConsole()->writeLine("### " . $synchro->getName() . " ###", ColorInterface::BLACK, ColorInterface::WHITE);
            $this->getConsole()->writeLine(date_format(date_create(), 'd/m/Y H:i:s'), ColorInterface::NORMAL);
            $this->getConsole()->writeLine();

            $result = $this->synchroService->runSynchro($synchro);

            if ($exception = $result->getException()) {
                $this->getConsole()->writeLine("Une erreur a été rencontrée!", ColorInterface::RED);
                $this->getConsole()->writeLine($exception->getMessage(), ColorInterface::RED);
                if ($previous = $exception->getPrevious()) {
                    $this->getConsole()->writeLine($previous->getMessage(), ColorInterface::RED);
                }
                $this->getConsole()->writeLine($exception->getTraceAsString(), ColorInterface::GRAY);
            }
            else {
                $this->getConsole()->writeLine((string)$result);
                $this->getConsole()->writeLine("Synchro réalisé avec succès.", ColorInterface::GREEN);
            }

            $startDate = $result->getStartDate();
            $endDate = $result->getEndDate();
            $diffDate = date_diff($startDate, $endDate);

            $this->getConsole()->writeLine($diffDate->format('%i min %s sec.'), ColorInterface::NORMAL);
            $this->getConsole()->writeLine();
        }
    }

    /**
     * @throws Exception
     */
    public function runImportAction()
    {
        /** @var Request $request */
        $request = $this->getRequest();

        //Passage de code à probablement concaténer en un (redonant au moins 3 fois)
        $all = $request->getParam('all') || $request->getParam('a');
        $name = $request->getParam('name');

        if ($all && $name !== null) {
            $this->getConsole()->writeLine("Les paramètres 'all' et 'name' sont antinomiques", ColorInterface::RED);
        }

        if (!$all && $name === null) {
            $all = TRUE;
        }

        $this->getConsole()->writeLine();
        $this->getConsole()->writeLine("######################## IMPORTS ########################", ColorInterface::BLUE);
        $this->getConsole()->writeLine();

        try {
            if ($all) {
                $imports = $this->importService->getImports();
            } else {
                try {
                    $import = $this->importService->getImportByName($name);
                    $imports = [$import];
                }
                catch (NotFoundException $exception) {
                    $imports = [];
                    $this->getConsole()->writeLine($exception->getMessage(), ColorInterface::RED);
                }
            }
        } catch (Exception $exception) {
            $this->getConsole()->writeLine("Une erreur imprévue est survenue durant l'import !", ColorInterface::RED);
            throw $exception;
        }

        $this->runImports($imports);

        $this->getConsole()->writeLine("");
    }

    /**
     * @throws Exception
     */
    public function runSynchroAction()
    {
        /** @var Request $request */
        $request = $this->getRequest();

        //Passage de code à probablement concaténer en un (redonant au moins 3 fois)
        $all = $request->getParam('all') || $request->getParam('a');
        $name = $request->getParam('name');

        if ($all && $name !== null) {
            $this->getConsole()->writeLine("Les paramètres 'all' et 'name' sont antinomiques", ColorInterface::RED);
        }

        if (!$all && $name === null) {
            $all = TRUE;
        }

        $this->getConsole()->writeLine();
        $this->getConsole()->writeLine("######################## SYNCHROS ########################", ColorInterface::BLUE);
        $this->getConsole()->writeLine();

        try {
            if ($all) {
                $synchros = $this->synchroService->getSynchros();
            } else {
                try {
                    $synchro = $this->synchroService->getSynchroByName($name);
                    $synchros = [$synchro];
                }
                catch (NotFoundException $exception) {
                    $synchros = [];
                    $this->getConsole()->writeLine($exception->getMessage(), ColorInterface::RED);
                }
            }
        } catch (Exception $exception) {
            $this->getConsole()->writeLine("Une erreur imprévue est survenue durant la synchro !", ColorInterface::RED);
            throw $exception;
        }

        $this->runSynchros($synchros);

        $this->getConsole()->writeLine("");
    }

}