<?php

namespace UnicaenDbImport\Domain;

use Assert\AssertionChain;
use Assert\InvalidArgumentException;
use Doctrine\DBAL\Connection as DbConnection;
use stdClass;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\Config\ConfigException;
use UnicaenDbImport\Connection\ApiConnection;
use Zend\Filter\FilterInterface;
use Zend\Stdlib\Parameters;

class Source implements SourceInterface
{
    protected $config;

    /**
     * @var array
     */
    protected $columns = [];

    /**
     * @var array Le résultat de la requête associée à la source ('select' ou 'table')
     */
    protected $data = [];

    /**
     * @param array $config
     * @return static
     */
    static public function fromConfig(array $config)
    {
        $inst = new static($config);

        try {
            $inst->validateConfig();
        } catch (\Exception $e) {
            if (! $e instanceof ConfigException) {
                $e = new ConfigException($e->getMessage(), 0, $e);
            }
            throw $e;
        }

        return $inst;
    }

    /**
     * Source constructor.
     *
     * @param array $config
     */
    protected function __construct(array $config)
    {
        $this->config = new Parameters($config);
    }

    /**
     * {@inheritDoc}
     */
    public function getName()
    {
        return $this->config->get('name');
    }

    /**
     * {@inheritDoc}
     */
    public function getTable()
    {
        return $this->config->get('table');
    }

    /**
     * {@inheritDoc}
     */
    public function getSelect()
    {
        return $this->config->get('select');
    }

    /**
     * {@inheritDoc}
     */
    public function getConnection()
    {
        return $this->config->get('connection');
    }

    /**
     * {@inheritDoc}
     */
    public function getSourceCodeColumn()
    {
        return strtoupper($this->config->get('source_code_column'));
    }

    /**
     * {@inheritDoc}
     */
    public function getColumns()
    {
        return array_map('strtoupper',$this->columns);
    }

    /**
     * {@inheritDoc}
     */
    public function getColumnNameFilter() : FilterInterface
    {
        return $this->config->get('column_name_filter');
    }

    /**
     * {@inheritDoc}
     */
    public function setColumns(array $columns)
    {
        $this->columns = $this->normalizeColumns($columns);
    }

    /**
     * {@inheritDoc}
     */
    public function getData(): array
    {
        return $this->data;
    }

    /**
     * {@inheritDoc}
     */
    public function setData(array $data)
    {
        $this->data = $this->normalizeData($data);
    }

    /**
     * {@inheritDoc}
     */
    public function getPageSize()
    {
        return (int) $this->config->get('page_size');
    }

    protected function validateConfig()
    {
        (new AssertionChain($this->config->get($key = 'name'), "Une string non vide est requise pour la clé suivante: $key"))
            ->notEmpty()
            ->string();

        if ($table = $this->config->get($key = 'table'))
            (new AssertionChain($table, "Une string non vide est requise pour la clé suivante: $key"))
                ->notEmpty()
                ->string();

        if ($select = $this->config->get($key = 'select'))
            (new AssertionChain($select, "Une string non vide est requise pour la clé suivante: $key"))
                ->notEmpty()
                ->string();

        if ($table && $select) {
            throw ConfigException::exclusiveKeys(['table', 'select']);
        }
        if (!$table && !$select) {
            throw ConfigException::atLeastOneKey(['table', 'select']);
        }

        $conn = $this->config->get($key = 'connection');
        if (! $conn instanceof DbConnection && ! $conn instanceof ApiConnection) {
            throw new InvalidArgumentException(
                "Une instance de " . DbConnection::class . " ou de " . ApiConnection::class . " est requise pour la clé suivante: $key",
                0, null, null
            );
        }

        (new AssertionChain($this->config->get($key = 'source_code_column'), "Une string non vide est requise pour la clé suivante: $key"))
            ->notEmpty()
            ->string();
    }

    /**
     * Normalise et valide les colonnes des données sources.
     *
     * @param array $columns
     * @return array
     */
    private function normalizeColumns(array $columns)
    {
        $sourceCodeColumn = $this->getSourceCodeColumn();

        // application du filtre de transformation
        $columnNameFilter = $this->getColumnNameFilter();
        $columns = array_map([$columnNameFilter, 'filter'], $columns);

        // recherche puis retrait de la "source code column"
        $index = array_search($sourceCodeColumn, $columns);
        if ($index === FALSE) {
            throw new RuntimeException(
                "La colonne de référence '$sourceCodeColumn' est introuvable dans les colonnes spécifiées");
        }
        unset($columns[$index]);

        // la source ne peut pas comporter de colonne nommée comme l'une des colonnes d'historique
        foreach (DestinationInterface::HISTO_COLUMNS as $col) {
            if ($found = in_array($col, $columns)) {
                throw new RuntimeException(sprintf(
                    "La source '%s' ne peut pas comporter de colonne nommée '%s'",
                    $this->getName(),
                    $col
                ));
            }
        }

        if (empty($columns)) {
            throw new RuntimeException(sprintf(
                "La source '%s' doit comporter au moins une colonne *en plus* de la colonne de référence '%s'",
                $this->getName(),
                $sourceCodeColumn
            ));
        }

        return $columns;
    }

    /**
     * Normalise les données sources : transforme les éventuels `stdClass` en `array`.
     *
     * @param array $data
     * @return array
     */
    private function normalizeData(array $data)
    {
        if (empty($data)) {
            return $data;
        }

        $first = reset($data);
        if (is_array($first)) {
            return $data;
        }
        if (! $first instanceof stdClass) {
            throw new RuntimeException(
                "Impossible de normaliser les données sources car leur type est inconnu (ni array ni stdClass)");
        }

        array_walk($data, function(&$value) {
            $value = get_object_vars($value);
        });

        return $data;
    }
}