<?php

namespace UnicaenDbImport\Service;

use Assert\Assertion;
use Assert\AssertionFailedException;
use Doctrine\DBAL\DBALException;
use Exception;
use stdClass;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\CodeGenerator\CodeGenerator;
use UnicaenDbImport\Domain\DestinationInterface;
use UnicaenDbImport\Domain\ResultInterface;
use UnicaenDbImport\Domain\SourceInterface;
use UnicaenDbImport\QueryExecutor;
use Zend\Filter\Callback;
use Zend\Filter\FilterInterface;
use Zend\Log\Logger;
use Zend\Log\LoggerInterface;
use Zend\Log\Writer\Stream;

class DatabaseService
{
    /**
     * @var CodeGeneratorPluginManager
     */
    protected $codeGeneratorPluginManager;

    /**
     * @var array
     */
    protected $codeGeneratorsConfig;

    /**
     * @var CodeGenerator
     */
    protected $codeGenerator;

    /**
     * @var CodeGenerator[]
     */
    protected $cachedCodeGenerators = [];

    /**
     * @var QueryExecutor
     */
    protected $queryExecutor;

    /**
     * @var SourceInterface $source
     */
    protected $source;

    /**
     * @var DestinationInterface $destination
     */
    protected $destination;

    /**
     * DatabaseHelper constructor.
     *
     * @param CodeGeneratorPluginManager $codeGeneratorPluginManager
     * @param array                      $codeGeneratorsConfig
     * @param QueryExecutor              $queryExecutor
     */
    public function __construct(
        CodeGeneratorPluginManager $codeGeneratorPluginManager,
        array $codeGeneratorsConfig,
        QueryExecutor $queryExecutor)
    {
        $this->codeGeneratorPluginManager = $codeGeneratorPluginManager;
        $this->codeGeneratorsConfig = $codeGeneratorsConfig;
        $this->queryExecutor = $queryExecutor;
    }

    /**
     * @param SourceInterface $source
     */
    public function setSource(SourceInterface $source): void
    {
        $this->source = $source;
    }

    /**
     * @param DestinationInterface $destination
     */
    public function setDestination(DestinationInterface $destination): void
    {
        $this->destination = $destination;

        $this->initCodeGenerator();
    }

    /**
     * @return CodeGenerator
     */
    public function initCodeGenerator()
    {
        if ($this->destination === null) {
            throw new RuntimeException("Une destination doit être spécifiée avant l'appel de cette méthode");
        }

        try {
            $platformClass = get_class($this->destination->getConnection()->getDatabasePlatform());
        } catch (DBALException $e) {
            throw new RuntimeException(
                "Impossible de déterminer la plateforme de la base de données destination", null, $e);
        }

        if (! isset($this->cachedCodeGenerators[$platformClass])) {
            try {
                Assertion::keyExists($this->codeGeneratorsConfig, $platformClass);
            } catch (AssertionFailedException $e) {
                throw new RuntimeException(
                    "Aucun CodeGenerator configuré pour la plateforme de base de données suivante : " . $platformClass, null, $e);
            }

            $codeGeneratorClass = $this->codeGeneratorsConfig[$platformClass];
            $codeGenerator = $this->codeGeneratorPluginManager->get($codeGeneratorClass);

            $this->cachedCodeGenerators[$platformClass] = $codeGenerator;
        }

        $this->codeGenerator = $this->cachedCodeGenerators[$platformClass];

        return $this->codeGenerator;
    }

    public function validateSourceTable()
    {
        $connection = $this->source->getConnection();
        $table = $this->source->getTable();

        // test if table exists
        $sql = $this->codeGenerator->generateSQLForTableExistenceCheck($table);
        $result = $this->queryExecutor->fetchAll($sql, $connection);
        $exists = $this->codeGenerator->convertTableExistenceCheckResultToBoolean($result);
        if (!$exists) {
            throw new RuntimeException("La table source '$table' est introuvable. ");
        }
    }

    public function validateDestinationTable()
    {
        $connection = $this->destination->getConnection();
        $table = $this->destination->getTable();

        // test if table exists
        $sql = $this->codeGenerator->generateSQLForTableExistenceCheck($table);
        $result = $this->queryExecutor->fetchAll($sql, $connection);
        $exists = $this->codeGenerator->convertTableExistenceCheckResultToBoolean($result);
        if (!$exists) {
            throw new RuntimeException("La table destination '$table' est introuvable. ");
        }

        // histo columns validation
        $sql = $this->codeGenerator->generateSQLForHistoColumnsValidation($table);
        $result = $this->queryExecutor->fetchAll($sql, $connection);
        $exception = $this->codeGenerator->convertHistoColumnsValidationBadResultToException($table, $result);
        if ($exception !== null) {
            throw new RuntimeException("La table '$table' n'est pas une table de destination valide. ", 0, $exception);
        }

        // source columns validation
        $sql = $this->codeGenerator->generateSQLForSourceColumnsValidation($table);
        $result = $this->queryExecutor->fetchAll($sql, $connection);
        $exception = $this->codeGenerator->convertSourceColumnsValidationBadResultToException($table, $result);
        if ($exception !== null) {
            throw new RuntimeException("La table '$table' n'est pas une table de destination valide. ", 0, $exception);
        }
    }

    /**
     *
     */
    public function createImportLogTableIfNotExists()
    {
        $connection = $this->destination->getConnection();
        $importLogTable = $this->destination->getLogTable();

        // test if table exists
        $sql = $this->codeGenerator->generateSQLForTableExistenceCheck($importLogTable);
        $result = $this->queryExecutor->fetchAll($sql, $connection);
        $exists = $this->codeGenerator->convertTableExistenceCheckResultToBoolean($result);
        if ($exists) {
            return;
        }

        // create table
        $sql = $this->codeGenerator->generateSQLForImportLogTableCreation($this->destination->getLogTable());
        $this->queryExecutor->exec($sql, $connection);
    }

    public function checkIntermediateTableNotExists($intermediateTable)
    {
        $sql = $this->codeGenerator->generateSQLForTableExistenceCheck($intermediateTable);
        $result = $this->queryExecutor->fetchAll($sql, $this->destination->getConnection());
        $exists = $this->codeGenerator->convertTableExistenceCheckResultToBoolean($result);
        if ($exists) {
            throw new RuntimeException(
                "Une table $intermediateTable existe déjà. Veuillez la supprimer ou alors spécifier dans la config " .
                "le nom à utiliser pour générer la table intermédiaire nécessaire à l'import dans la table " .
                $this->destination->getTable());
        }
    }

    public function createIntermediateTable($intermediateTable)
    {
        $sql = $this->codeGenerator->generateSQLForIntermediateTableCreation($this->source, $this->destination, $intermediateTable);

        $this->queryExecutor->exec($sql, $this->destination->getConnection());
    }

    public function dropIntermediateTable($intermediateTable)
    {
        $sql = $this->codeGenerator->generateSQLForIntermmediateTableDrop($intermediateTable);

        $this->queryExecutor->exec($sql, $this->destination->getConnection());
    }

    public function truncateDestinationTable()
    {
        $sql = $this->codeGenerator->generateSQLForTruncateTable($this->destination->getTable());

        $this->queryExecutor->exec($sql, $this->destination->getConnection());
    }

    /**
     * @return int
     * @throws Exception
     */
    public function populateDestinationTableFromSource()
    {
        $destinationTable = $this->destination->getTable();

        return $this->populateTableFromSource($destinationTable);
    }

    /**
     * @param $tableName
     * @return int
     */
    public function populateTableFromSource(string $tableName)
    {
        $columns = $this->source->getColumns();
        $sourceCodeColumn = $this->source->getSourceCodeColumn();
        $columns = array_merge([$sourceCodeColumn], $columns);

        $rows = $this->source->getData();

        $insertsSQL = '';
        foreach ($rows as $row) {
            $preparedRow = $this->prepareSourceDataForDestination($row);
            $columnsValues = [];
            foreach ($columns as $column) {
//                $columnExistsInData = array_key_exists(strtolower($column), $preparedRow) || array_key_exists(strtoupper($column), $preparedRow);
//                if (!$columnExistsInData) {
//                    continue;
//                }
                $columnsValues[$column] = $preparedRow[$column] ?? null;
            }
            $sql = $this->codeGenerator->generateSQLForInsertIntoTable($tableName, $columnsValues);

            $insertsSQL .= $sql . PHP_EOL;
        }

        // insert into destination
        $this->queryExecutor->exec($insertsSQL, $this->destination->getConnection()); //Exécution de la requêtes (inserts) dans la base destination

        return count($rows);
    }

    private function prepareSourceDataForDestination(array $row)
    {
        $columnNameFilter = $this->source->getColumnNameFilter();

        $newKeys = array_map([$columnNameFilter, 'filter'], array_keys($row));
        $row = array_combine($newKeys, $row);

        return $row;
    }

    /**
     * @param string $intermediateTable
     */
    public function recreateDiffView($intermediateTable)
    {
        try {
            $sql = $this->codeGenerator->generateSQLForDiffViewCreation($this->source, $this->destination, $intermediateTable);
        } catch (DBALException $e) {
            throw new RuntimeException("Erreur rencontrées lors de la création de la vue diff", null, $e);
        }

        $this->queryExecutor->exec($sql, $this->destination->getConnection());
    }

    /**
     * @return array
     */
    public function executeDiffViewRequest()
    {
        // interrogation de la vue V_DIFF_
        $sql = $this->codeGenerator->generateSQLForDiffViewSelect($this->source, $this->destination);
        $rows = $this->queryExecutor->fetchAll($sql, $this->destination->getConnection());

        return $rows;
    }

    /**
     * @param array $rows
     */
    public function updateDestinationFromDiffViewRequestResult(array $rows)
    {
        // à partir du diff, génération des instructions SQL d'insert et d'update nécessaires pour mettre à jour la table destination
        $sql = $this->codeGenerator->generateSQLForDestinationUpdateFromDiffViewRequestResult($rows, $this->source, $this->destination);
        if ($sql) {
            $this->queryExecutor->exec($sql, $this->destination->getConnection());
        }
    }

    /**
     * Requête la Source pour obtenir les données sources.
     *
     * @return array
     */
    public function fetchSource()
    {
        $selectSql = $this->codeGenerator->generateSQLForSelectFromSource($this->source);

        return $this->queryExecutor->fetchAll($selectSql, $this->source->getConnection());
    }

    /**
     * @param ResultInterface $result
     */
    public function saveResultToLogTable(ResultInterface $result)
    {
        $importLogTable = $this->destination->getLogTable();

        $sql = $this->codeGenerator->generateSQLForInsertResultIntoLogTable($result, $importLogTable);

        $this->queryExecutor->exec($sql, $this->destination->getConnection());
    }
}