<?php

namespace UnicaenDbImport\Service;

use Doctrine\DBAL\Connection as DbConnection;
use Exception;
use stdClass;
use UnicaenApp\Exception\RuntimeException;
use UnicaenDbImport\Connection\ApiConnection;
use UnicaenDbImport\Domain\DestinationInterface;
use UnicaenDbImport\Domain\Exception\ConnectionException;
use UnicaenDbImport\Domain\ResultInterface;
use UnicaenDbImport\Domain\SourceInterface;
use Zend\Log\Logger;
use Zend\Log\LoggerInterface;
use Zend\Log\Writer\Stream;

class FacadeService
{
    /**
     * @var DatabaseService
     */
    private $databaseService;

    /**
     * @var ApiService
     */
    private $apiService;

    /**
     * @var SourceInterface $source
     */
    protected $source;

    /**
     * @var DestinationInterface $destination
     */
    protected $destination;

    /**
     * FacadeService constructor.
     *
     * @param DatabaseService $databaseService
     * @param ApiService      $apiService
     */
    public function __construct(DatabaseService $databaseService, ApiService $apiService)
    {
        $this->databaseService = $databaseService;
        $this->apiService = $apiService;
    }

    /**
     * @param SourceInterface $source
     */
    public function setSource(SourceInterface $source): void
    {
        $this->source = $source;
        $this->databaseService->setSource($this->source);
        $this->apiService->setSource($this->source);
    }

    /**
     * @param DestinationInterface $destination
     */
    public function setDestination(DestinationInterface $destination): void
    {
        $this->destination = $destination;
        $this->databaseService->setDestination($this->destination);
        $this->apiService->setDestination($this->destination);
    }

    public function validateSource()
    {
        $connection = $this->source->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                if (! $this->requiresIntermediateTable()) {
                    $this->databaseService->validateSourceTable();
                }
                break;
            case $connection instanceof ApiConnection:
                // Ex:
                // $this->apiService->validateSourceApi();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    public function validateDestination()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->validateDestinationTable();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    public function createImportLog()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->createImportLogTableIfNotExists();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    /**
     *
     */
    public function clearDestination()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->truncateDestinationTable();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    /**
     * @return int
     * @throws Exception
     */
    public function importToDestination()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                return $this->databaseService->populateDestinationTableFromSource();
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    /**
     * @return int
     */
    public function populateIntermediateTableFromSource()
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $intermediateTable = $this->destination->getIntermediateTable();
                return $this->databaseService->populateTableFromSource($intermediateTable);
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }

    /**
     * @return array
     */
    public function synchronizeDestination()
    {
        $intermediateTable = null;

        if ($this->requiresIntermediateTable()) {
            $intermediateTable = $this->destination->getIntermediateTable();
            if ($this->destination->getIntermediateTableAutoDrop()) {
                $this->databaseService->dropIntermediateTable($intermediateTable);
            } else {
                $this->databaseService->checkIntermediateTableNotExists($intermediateTable);
            }
            $this->databaseService->createIntermediateTable($intermediateTable);
            $this->databaseService->populateTableFromSource($intermediateTable);
        }

        $this->databaseService->recreateDiffView($intermediateTable);
        $diffViewRequestResult = $this->databaseService->executeDiffViewRequest();
        $this->databaseService->updateDestinationFromDiffViewRequestResult($diffViewRequestResult);

        if ($this->requiresIntermediateTable()) {
            $this->databaseService->dropIntermediateTable($intermediateTable);
        }

        return $diffViewRequestResult;
    }

    /**
     * @return bool
     */
    private function requiresIntermediateTable()
    {
        $isSameConnectionForSourceAndDestination = $this->source->getConnection() === $this->destination->getConnection();

        return ! $isSameConnectionForSourceAndDestination;
    }

    /**
     * Requête la Source pour déterminer la liste des colonnes et les données sources elles-mêmes.
     * Ces informations sont ensuite disponibles dans la Source en question.
     *
     * @see SourceInterface::getColumns()
     * @see SourceInterface::getData()
     */
    public function fetchSource()
    {
        if (count($this->source->getData()) === 0) {

            /** @var ApiConnection|DbConnection $connection */
            $connection = $this->source->getConnection();

            switch (true) {
                case $connection instanceof ApiConnection:
                    $this->apiService->setLogger($this->createLogger());
                    $this->apiService->setConfig($connection->toArray());
                    $data = $this->apiService->callApi($connection, $this->source);
                    break;

                case $connection instanceof DbConnection:
                    $data = $this->databaseService->fetchSource();
                    break;

                default:
                    throw new RuntimeException("Type de connexion source inattendue");
                    break;
            }

            $columns = $this->extractColumnsFromSourceFetchResults($data);

            $this->source->setData($data);
            $this->source->setColumns($columns);
        }
    }

    /**
     * @param array $rows
     * @return string[]
     */
    private function extractColumnsFromSourceFetchResults(array $rows)
    {
        if (empty($rows)) {
            throw new RuntimeException(
                "Impossible de déterminer la liste des colonnes à partir des données sources car celles-ci sont vides");
        }

        $first = current($rows);

        if ($first instanceof stdClass) {
            $columns = array_keys(get_object_vars($first));
        } elseif (is_array($first)) {
            $columns = array_keys($first);
        } else {
            throw new RuntimeException(
                "Impossible de déterminer la liste des colonnes à partir des données sources car leur type est inconnu (ni array ni stdClass)");
        }

        return $columns;
    }

    /**
     * @return LoggerInterface
     */
    private function createLogger()
    {
        $logger = new Logger();
        $logger->addWriter(new Stream('php://output'));

        return $logger;
    }

    /**
     * @param ResultInterface $result
     */
    public function saveResultToImportLog(ResultInterface $result)
    {
        $connection = $this->destination->getConnection();

        switch (true) {
            case $connection instanceof DbConnection:
                $this->databaseService->saveResultToLogTable($result);
                break;
            default:
                throw ConnectionException::unexpected($connection);
                break;
        }
    }
}