<?php

namespace UnicaenDbImport\Service;

use Exception;
use UnicaenDbImport\Config\Config;
use UnicaenDbImport\Domain\Exception\NotFoundException;
use UnicaenDbImport\Domain\ImportInterface;
use UnicaenDbImport\Domain\ImportResult;

class ImportService
{
    /**
     * @var Config
     */
    private $config;

    /**
     * @var FacadeService
     */
    private $facadeService;

    /**
     * ImportService constructor.
     *
     * @param FacadeService $facadeService
     * @param Config        $config
     */
    public function __construct(
        FacadeService $facadeService,
        Config $config)
    {
        $this->facadeService = $facadeService;
        $this->config = $config;
    }

    /**
     * Retourne tous les imports (bruts) issus de la config.
     *
     * @return ImportInterface[]
     */
    public function getImports()
    {
        return $this->config->getImports();
    }

    /**
     * Lance tous les imports.
     *
     * @return ImportResult[] Agrégat des résultats des l'exécutions des imports
     */
    public function runAllImports()
    {
        $imports = $this->getImports();
        $results = [];

        foreach ($imports as $import) {
            $results[$import->getName()] = $this->runImport($import);
        }

        return $results;
    }

    /**
     * Retourne l'import dont le nom est spécifié.
     *
     * @param string $name Nom de l'import (name)
     * @return ImportInterface Import trouvé
     *
     * @throws NotFoundException Import introuvable
     */
    public function getImportByName($name)
    {
        $import = $this->config->getImport($name);
        if ($import === null) {
            throw NotFoundException::importByName($name);
        }

        return $import;
    }

    /**
     * Lance l'import spécifié (import brut).
     *
     * @param ImportInterface $import
     * @return ImportResult Résultat de l'exécution de l'import
     */
    public function runImport(ImportInterface $import)
    {
        $source = $import->getSource();
        $destination = $import->getDestination();

        $result = new ImportResult();
        $result->setImport($import);

        $startDate = date_create('now');

        $this->facadeService->setSource($source);
        $this->facadeService->setDestination($destination);

        try {
            $this->facadeService->createImportLog();
            $this->facadeService->validateDestination();
            $this->facadeService->fetchSource();
            $this->facadeService->clearDestination();
            $count = $this->facadeService->importToDestination();

            $result->setDestinationTablePopulateResult($count);

        } catch (Exception $e) {
            $result->setException($e);
        }

        $result->setStartDate($startDate);
        $result->setEndDate();

        $this->facadeService->saveResultToImportLog($result);

        return $result;
    }
}