<?php


namespace UnicaenDbImport\Service;

use Exception;
use UnicaenDbImport\Config\Config;
use UnicaenDbImport\Domain\Exception\NotFoundException;
use UnicaenDbImport\Domain\ImportInterface;
use UnicaenDbImport\Domain\SynchroInterface;
use UnicaenDbImport\Domain\SynchroResult;

class SynchroService
{
    /**
     * @var Config
     */
    private $config;

    /**
     * @var FacadeService
     */
    private $facadeService;

    /**
     * SynchroService constructor.
     *
     * @param FacadeService $facadeService
     * @param Config        $config
     */
    public function __construct(FacadeService $facadeService, Config $config)
    {
        $this->facadeService = $facadeService;
        $this->config = $config;
    }


    /**
     * Retourne toutes les synchros issues de la config.
     *
     * @return SynchroInterface[]
     */
    public function getSynchros()
    {
        return $this->config->getSynchros();
    }

    /**
     * Lance tous les synchros.
     *
     * @return SynchroResult[] Agrégat des résultats des l'exécutions des synchros
     */
    public function runAllSynchros()
    {
        $synchros = $this->getSynchros();
        $results = [];

        foreach ($synchros as $synchro) {
            $results[$synchro->getName()] = $this->runSynchro($synchro);
        }

        return $results;
    }

    /**
     * Retourne la synchro dont le nom est spécifié.
     *
     * @param string $name Nom de la synchro (name)
     * @return ImportInterface Synchro trouvée
     *
     * @throws NotFoundException Synchro introuvable
     */
    public function getSynchroByName($name)
    {
        $synchro = $this->config->getSynchro($name);
        if ($synchro === null) {
            throw NotFoundException::synchroByName($name);
        }

        return $synchro;
    }

    /**
     * Lance la synchro spécifiée.
     *
     * @param SynchroInterface $synchro
     * @return SynchroResult Résultat de l'exécution de la synchro
     */
    public function runSynchro(SynchroInterface $synchro)
    {
        $source = $synchro->getSource();
        $destination = $synchro->getDestination();

        $result = new SynchroResult();
        $result->setSynchro($synchro);

        $this->facadeService->setSource($source);
        $this->facadeService->setDestination($destination);

        $intermediateTable = null;

        $startDate = date_create('now');

        try {
            $this->facadeService->createImportLog();
            $this->facadeService->validateDestination();
            $this->facadeService->validateSource();
            $this->facadeService->fetchSource();
            $diffViewRequestResult = $this->facadeService->synchronizeDestination();

            $result->setDiffViewRequestResult($diffViewRequestResult);

        } catch (Exception $e) {
            $result->setException($e);
        }

        $result->setStartDate($startDate);
        $result->setEndDate();

        $this->facadeService->saveResultToImportLog($result);

        return $result;
    }


}